# -*- coding: utf-8 -*-
"""

    Agatta: Three-item analysis Python package
    Contact: Valentin Rineau - valentin.rineau@sorbonne-universite.fr

    Agatta is a set of tools in the cladistic framework to perform
    three-item analysis and associated operations in cladistics in python.

    https://github.com/vrineau/agatta

    This code is under license GNU GPLv3

"""

from random import choice
from .ini import taxa_extraction
import os
import sys
import time
import platform
from ete3 import Tree


def search_pipeline(path_infile, software_path=False, software="paup",
                    prefix=False):
    """
    Function to run automatically a nexus file, a tnt file, or a text file on
    PAUP, TNT, WQFM, or wTREE-QMC respectivelly. The software to be used must 
    be installed and accessible.

    Parameters
    ----------
    path_infile : string
        path of the input file.
    software_path : string
        path of the software chosen: paup, tnt, wqfm, wtree-qmc. Not necessary 
        if paup is installed in a windows os.
    software : string, optional
        paup, tnt, wqfm, wtree-qmc. The default is "paup".

    Returns
    -------
    Nothing. Run the analysis with specified file and software

    """

    print("Running analysis on " + software + " software")
    print("==================================================================")
    start = time.time()

    if software_path:
        software_path = os.path.expanduser(software_path)
        if not os.path.isfile(software_path):
            print("ERROR: '" + software_path
                           + "' folder does not exist.\nOperation aborted.")
            sys.exit(1)

    ostype = platform.system()  # os detection

    if ostype == "Windows":
        beginline = "paup"
        #endline = ">NUL"
    else:
        beginline = software_path
        endline = "> /dev/null"


    if software == "tnt":
        print("WARNING: 3ia analysis using TNT is still under development.\n"
              + "Searches using heuristics and fractional weights are not "
              + "optimized.")
        if " " in path_infile:
            print("ERROR: TNT software doesn't allow paths with spaces.\n" +
                  "The path \"" + path_infile + "\" contains spaces.")
            sys.exit(1)

        os.system(software_path + " proc "+path_infile + endline)

    if software == "wqfm":
        os.system("java -jar \"" + software_path + "\" -i \"" +
                  path_infile + "\" -o \"" + prefix + ".tre\"")

    elif software == "wtree-qmc":
        os.system("\"" + software_path + "\" -i \"" +
                  path_infile + "\" --quartets -o \"" + prefix + ".tre\"")

    elif software == "paup":

        os.system(beginline + " -n \"" + path_infile + "\"")

    print("==================================================================")
    print("Analysis done")

    end = time.time()
    time_cptr = time.strftime('%H:%M:%S', time.gmtime(end - start))
    print("elapsed time (3-item analysis): {}".format(time_cptr))


def rear_taxa(tree1):
    """
    Takes as argument an ete3 Tree and outputs the same tree with pruning
    and regrafting of a leaf.

    Parameters
    ----------
    tree1 : ete3.Tree
        Rooted tree.

    Returns
    -------
    tree2 : ete3.Tree
        Rearranged tree.
    leaf_node : ete3.Tree
        leaf node pruned and regrafted on the new tree

    """

    character_dict = {}
    character_dict[tree1.copy(method="newick")] = str(1)
    tree_copy_test = tree1.copy(method="newick")
    rf = 0  # similarity score between old and new tree

    while rf == 0:  # while trees are the same
        tree2 = tree_copy_test.copy(method="newick")
        rand_node = Tree()
        rand_taxa = ""

        # choice of pruned leaf and regraft location
        # check if pruned and regraft location are not the same
        while rand_node.name == rand_taxa:
            # location where to regraft the leaf
            rand_node = choice(
                [node for node in tree2.traverse(strategy="postorder")])
            # taxon randomly chosen to be  pruned and regrafted
            rand_taxa = choice(
                [taxa for taxa in taxa_extraction(character_dict)])

        regraft_node = rand_node.copy(method="newick")
        tree2.search_nodes(name=rand_taxa)[0].get_ancestors()[0].delete()
        tree2.search_nodes(name=rand_taxa)[0].delete()  # del taxon

        if regraft_node.search_nodes(name=rand_taxa):
            regraft_node.search_nodes(
                name=rand_taxa)[0].get_ancestors()[0].delete()
            regraft_node.search_nodes(name=rand_taxa)[0].delete()  # del taxon

        if len([leaf for leaf in rand_node.iter_leaves()]) == 1:
            leaf_name = rand_node.get_leaves()[0].name
            rand_node.get_leaves()[0].name = "new_ancestor"
            new_ancestor = rand_node.get_leaves()[0]
            new_ancestor.add_child(name=leaf_name)
            new_ancestor.add_child(name=rand_taxa)

        else:
            for node in rand_node.get_children():  # node deletion
                node.detach()
            new_ancestor = rand_node.add_child(name="new_ancestor")  # regraft
            new_ancestor.add_child(regraft_node)
            rand_node.add_child(name=rand_taxa)

        for node in tree2.traverse(strategy="levelorder"):
            if len(node.get_children()) == 1:
                node.children[0].delete()
        rf, max_rf, common_leaves, parts_t1, parts_t2, set1, set2 = \
            tree2.robinson_foulds(tree_copy_test)

    leaf_node = tree2.search_nodes(name=rand_taxa)[0]

    return tree2, leaf_node


def tripletscore(triplet_dict, stree):
    """
    Compute the triplet score of a dichotomic tree for a branch and bound
    analysis.
    The triplet score is the sum of weights of the triplets compatibles with
    the tree

    Parameters
    ----------
    triplet_dict : dict
        Dictionary of triplets generated by main_tripdec().
    stree : ete3.Tree
        Rooted tree.

    Returns
    -------
    score : int, float or fraction
        Score of the tree. The type depends of the type of the triplet weights.

    """

    score = 0

    for triplet, FW in triplet_dict.items():
        tripin = list(triplet.in_taxa)
        a = stree.get_leaves_by_name(tripin[0])
        b = stree.get_leaves_by_name(tripin[1])
        c = stree.get_leaves_by_name(list(triplet.out_taxa)[0])

        if a and b and c:  # if the three leaves are in the tree

            common = stree.get_common_ancestor([a[0], b[0]])
            cc = common.get_leaves_by_name(list(triplet.out_taxa)[0])

            if cc:  # if triplet not compatible
                score += FW*2
            else:  # if triplet compatible
                score += FW

    return score


def triplet_check(triplet, stree):
    """
    check if triplet exists in tree.
    Returns true or false.

    """

    tripin = list(triplet.in_taxa)
    a = stree.get_leaves_by_name(tripin[0])
    b = stree.get_leaves_by_name(tripin[1])
    c = stree.get_leaves_by_name(list(triplet.out_taxa)[0])

    if a and b and c:  # if the three leaves are in the tree

        common = stree.get_common_ancestor([a[0], b[0]])
        cc = common.get_leaves_by_name(list(triplet.out_taxa)[0])

        if cc:  # if triplet not compatible
            score = False
        else:  # if triplet compatible
            score = True

    return score




# -*- coding: utf-8 -*-
"""

    Agatta: Three-item analysis Python package
    Contact: Valentin Rineau - valentin.rineau@sorbonne-universite.fr

    Agatta is a set of tools in the cladistic framework to perform
    three-item analysis and associated operations in cladistics in python.

    https://github.com/vrineau/agatta

    This code is under license GNU GPLv3

"""

from .ini import wrapper_character_extraction
from .ini import taxa_to_numbers
from .ini import infotree_checker
from fractions import Fraction
from itertools import combinations
from itertools import product
from multiprocessing import Pool
from multiprocessing import cpu_count
from collections import defaultdict
from functools import partial
from tqdm import tqdm
import os
import sys
import csv
import time
import uuid
import shutil
import pickle
import pandas as pd
from ete3 import Tree


class triplet():

    __slots__ = ("in_taxa", "out_taxa", "FW")

    def __init__(self, in_taxa=set(), out_taxa=set(), FW=Fraction()):

        self.in_taxa = set(in_taxa)  # taxa in the informative node
        self.out_taxa = set(out_taxa)  # sister group of the informative node
        self.FW = FW  # Triplet weight

    def __eq__(self, other_triplet):

        return (set(self.in_taxa) == set(other_triplet.in_taxa)) and (
            self.out_taxa == other_triplet.out_taxa)

    def __call__(self, in_taxa=set(), out_taxa=set(), FW=Fraction(),
                 parent_triplets=()):

        return self

    def __hash__(self):

        return hash(list(self.in_taxa)[0]) ^ hash(list(self.in_taxa)[1]) ^ \
            hash(list(self.out_taxa)[0])

    def __del__(self):

        del self

    def __repr__(self):

        return "({},({},{}))".format(str(list(self.out_taxa)[0]),
                                    str(sorted(list(self.in_taxa))[0]),
                                    str(sorted(list(self.in_taxa))[1]))


def del_replications(treerep, method="TMS", verbose=False):
    """
    Remove all repeated leaves from a single rooted tree according to the
    free-paralogy subtree analysis. Two algorithms are currently implemented in
    Agatta:

        * The original algorithm of Nelson and Ladiges (1996) designed in the
          paradigm of cladistic biogeography: Free-paralogy Subtree analysis.

        * The algorithm of Rineau et al. (2022) designed to construct
          subtrees without repeated leaves while minimising the loss of
          information in terms of triplets: Triplet Maximisation Subtrees.
          This is the default mode.

          Nelson, G. J., & Ladiges, P. Y. (1996). Paralogy in cladistic
          biogeography and analysis of paralogy-free subtrees.
          American Museum novitates 3167.

          Rineau, V., Moncel, M-H., & Zeitoun, V. (2022). Revealing
          evolutionary patterns behind homogeneity: the case of the Paleolithic
          assemblages from Notarchirico (Southern Italy).

    Parameters
    ----------
    treerep : ete3.Tree
        One rooted tree.
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS". The default is "TMS".
    verbose : bool, optional
        Verbose mode if true. The default is False.

    Returns
    -------
    list
        A list of subtrees without repeated leaves.

    """

    def del_replications_node(treerep, method="TMS", verbose=False):
        """
            Function to remove a single repeated leaf.
            Used in del_replications.
        """

        sys.setrecursionlimit(10**7)  # can hit the recursion limit

        paralogs = dict()
        cardl = list(Tree.get_leaf_names(treerep))
        cards = set(Tree.get_leaf_names(treerep))
        cardr = [x for x in cards if cardl.count(x) > 1]  # list of rep. leaves
        treelistrep = []
        treelistnorep = []
        treelistverif = []

        if cardr:  # if repetitions
            nodelist = sorted([node for node in treerep.traverse(
                strategy="postorder") if not node.is_leaf()],
                key=lambda n: len(n))

            for e in nodelist:  # iterate from the smallest to the biggest node
                leafset = set(Tree.get_leaf_names(e))

                # analysis of the smallest node with repetitions
                if len(leafset) != len(Tree.get_leaf_names(e)):  # if rep.

                    # Identification if rep. is: apical, ortholog, paralog
                    # apical: node including only leaves
                    # ortholog: asymmetric node
                    # paralog: symmetric node
                    i = 0  # number of internal nodes connected to e
                    e.add_feature("FP", "main")
                    for child_node in e.get_children():
                        if not child_node.is_leaf():
                            i += 1
                            for cnode in child_node.traverse(
                                    strategy="postorder"):
                                cnode.add_feature("FP", i)  # label nodes save

                    for node in treerep.traverse():
                        try:
                            if node.FP:
                                pass
                        except AttributeError:
                            node.add_feature("FP", "out")

                    # special case: repeated leaf branched to a symmetric node
                    paralog_but_leaf = False

                    if i > 1:
                        special_tree = treerep.copy(method='cpickle')
                        for child in special_tree.search_nodes(
                                FP="main")[0].get_children():
                            if child.is_leaf:
                                if child.name in cardr:
                                    paralog_but_leaf = True
                                    if verbose:
                                        print("ortholog repetition (on "
                                              + "symmetric node):{}".format(
                                                  child.name))
                                    child.delete()

                    # special case: leaves on paralog
                    if paralog_but_leaf:
                        paralogs["main"] = special_tree

                    # apical
                    elif i == 0:
                        paralogs["main"] = treerep.copy(method='cpickle')
                        for l in set(cardr) & leafset:  # for each leaf
                            leaf1 = False
                            for delnode in paralogs["main"].search_nodes(
                                    FP="main")[0].get_leaves_by_name(name=l):
                                if not leaf1:
                                    leaf1 = True  # except the first
                                else:
                                    delnode.delete()  # delete leaves
                                    if verbose:
                                        print("apical repetition:{}".format(
                                            delnode.name))

                    # ortholog
                    elif i == 1:
                        paralogs["main"] = treerep.copy(method='cpickle')
                        kl = []

                        for l in set(cardr) & leafset:  # for each leaf

                            leaf1 = False

                            for sorted_nodes in sorted(
                                    [node for node in paralogs[
                                        "main"].search_nodes(
                                            FP="main")[0].traverse(
                                                strategy="postorder") if not
                                        node.is_leaf()], key=lambda n: len(n)):

                                for childnode in sorted_nodes.get_children():
                                    if childnode.is_leaf():
                                        if childnode.name == l:  # for each
                                            if not leaf1:  # rep. leaf
                                                leaf1 = True    # except first
                                            elif leaf1:
                                                kl.append(childnode)

                        for delnode in kl:
                            delnode.delete()  # delete leaves
                            if verbose:
                                print("ortholog repetition:{}".format(
                                    delnode.name))

                    # paralog
                    else:

                        j = 1

                        # Rineau's algorithm - Triplet Maximisation Subtrees
                        if method == "TMS":
                            paralogs["main"] = treerep.copy(method='cpickle')

                            for delnode in paralogs["main"].search_nodes(
                                   FP="main")[0].iter_descendants("postorder"):
                                if not delnode.is_leaf():
                                    delnode.delete()  # del internal node

                            while j != i+1:
                                paralogs[j] = treerep.copy(method='cpickle')

                                for delnode in paralogs[j].traverse():
                                    if (not delnode.is_leaf() and not
                                    delnode.FP == j):
                                        delnode.delete()  # del internal node

                                j += 1

                        # Nelson's algorithm - Free paralogy subtree analysis
                        elif method == "FPS":

                            while j != i+1:  # for each node brch to paralog

                                # subtree for each node branched to the paralog
                                paralogs[j] = treerep.copy(method='cpickle')

                                for delnode in paralogs[j].search_nodes(
                                        FP="main")[0].get_children():
                                    if not delnode.FP == j:
                                        delnode.detach()

                                paralogs[j].search_nodes(FP="main")[0].delete()
                                j += 1

                        if verbose:
                            print("paralog repetition. " + str(len(paralogs)) +
                                  " subtrees " +
                                  "generated from the main tree.")

                    break  # function only for one instance of repetition

            treelistverif = [p for p in paralogs.values()]

        else:  # if there is no repetition
            treelistnorep.append(treerep)

        # sort between trees with and without repetitions
        for t in treelistverif:
            if len(set(Tree.get_leaf_names(t))) > 2:
                if len(list(Tree.get_leaf_names(t))) != len(set(
                        Tree.get_leaf_names(t))):  # detect repetitions
                    treelistrep.append(t)
                else:
                    treelistnorep.append(t)

        # del features (main) in all trees for recursion
        treelistrepclearcopy = []
        for t in treelistrep:
            for node in t.traverse():
                node.del_feature('FP')
            treelistrepclearcopy.append(t.copy(method='cpickle'))

        return treelistrepclearcopy, treelistnorep

    listundone, listdone = del_replications_node(treerep, method)

    if listundone:  # if repetitions
        while listundone:  # while until listundone contains trees to analyse
            listundone1 = []
            listdone1 = []

            for treerep2 in listundone:
                LU, LD = del_replications_node(treerep2, method)
                listundone1 += LU
                listdone1 += LD

            # all trees analysed
            listundone = listundone1.copy()  # replace trees with rep. by list
            listdone += listdone1  # accumulation of finished trees in listdone

    # detect non-informative trees
    treelist = []
    for infotree in listdone:

        # detect and delete empty internal node (one single descendant)
        for node in infotree.traverse():
            if not node.is_leaf() and len(node.get_children()) == 1:
                node.get_children()[0].delete()

        # retain tree only if more than one internal node
        cdict, noninfo = infotree_checker({infotree:1}, verbose=False)
        if cdict.keys():
            infotree = list(cdict.keys())[0]
            infotree.ladderize()
            treelist.append(infotree)

    if verbose:  # verbose mode

        print("Output trees:")
        if treelist:
            for l in treelist:
                print(l.write(format=9))
        else:
            print("No informative tree")
        print("")

    return treelist


def del_replications_forest(character_dict, method="TMS",
                            prefix="agatta_del_replications", verbose=False):
    """
    Remove all repeated leaves from trees according to the
    free-paralogy subtree analysis. Two algorithms are currently implemented in
    Agatta:

        * The original algorithm of Nelson and Ladiges (1996) designed in the
          paradigm of cladistic biogeography.

        * The algorithm of Rineau et al. (2022) designed to construct
          subtrees without repeated leaves while minimising the loss of
          information in terms of triplets. This is the default mode.

          Nelson, G. J., & Ladiges, P. Y. (1996). Paralogy in cladistic
          biogeography and analysis of paralogy-free subtrees.
          American Museum novitates 3167.

          Rineau, V., Moncel, M-H., & Zeitoun, V. (2022). Revealing
          evolutionary patterns behind homogeneity: the case of the Paleolithic
          assemblages from Notarchirico (Southern Italy).

    Parameters
    ----------
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS". The default is "TMS".
    prefix : str, optional
        Prefix of the saving file. The default is False.
    verbose : bool, optional
        Verbose mode if true. The default is False.

    Returns
    -------
    tree_dict
        A dictionary of resulting subtrees in keys and their id as values.
        Id is a string. If a tree 2 results in a single subtree, the id of the
        subtree will be '[2]', otherwise if several subtrees are generated
        their respective id will be '[2.1]', '[2.2]', etc.

    """
    print("Managing polymorphism")
    tree_dict = dict()
    output_trees = 0


    if verbose:
        loopchar = character_dict.items()
    else:
        loopchar = tqdm(character_dict.items())
        
    if prefix:
        with open(prefix+".poly", "w") as logfile:
            logfile.write("Polymorphism management: {}\n".format(method))

    for treed, treeid in loopchar:

        if verbose:
            print("[" + str(treeid) + "] " + treed.write(format=9))
        
        if prefix:
            with open(prefix+".poly", "a") as logfile:
                logfile.write("[" + str(treeid) + "] " 
                              + treed.write(format=9) + "\n")

        treelist = del_replications(treed, method, verbose)

        if treelist:
            
            # if single tree obtained
            if len(treelist) == 1:
                tree_dict[treelist[0]] = treeid
                
                if prefix:
                    if len(treed) == len(treelist[0]):
                        with open(prefix+".poly", "a") as logfile:
                            logfile.write("[" + str(treeid) + "] " +
                                          "no repeated leaf\n")
                    else:
                        with open(prefix+".poly", "a") as logfile:
                            logfile.write("[" + str(treeid) + "] " +
                                          treelist[0].write(format=9) + "\n")
                        
            # if character cut into several free-paralogy subtrees
            else:
                
                # function to sort trees with their minimal state number 
                def sorttreelist(character):
                    """
                    function to sort trees with their min state nb
                    return empty list otherwise
                    """
                    statenb = []
                    
                    try: 
                        for n in character.traverse(strategy="preorder"):
                            if (n.is_leaf() == False and 
                                n.is_root() == False):
                                statenb.append(int(n.charstate_name))
                        
                        return min(statenb) 
                    
                    except AttributeError:
                        return 1000
                          
                #reorder list with state numbers if state numbers
                treelist.sort(key=sorttreelist)

                i = 1
                output_trees += len(treelist)
                for treel in treelist:
                    tree_dict[treel] = str(treeid) + "." + str(i)


                    if prefix:
                        with open(prefix+".poly", "a") as logfile:
                            logfile.write("[" + str(treeid) + "." + str(i) + 
                                          "] " + treel.write(format=9) + "\n")
                    i += 1
        else:
            if prefix:
                with open(prefix+".poly", "a") as logfile:
                    logfile.write("[" + str(treeid) + 
                                  "] no informative tree\n")

        if prefix:
            with open(prefix+".poly", "a") as logfile:
                logfile.write("\n")

    td, noninfo = infotree_checker(tree_dict)
    
    # message when non-informative characters
    for n in noninfo:
        print("Character {} non-informative after " 
              +"polymorphism management".format(n))

    if output_trees != 0:
        print("Polymorphism removed: "
              "{} polymorphism free informative characters "
              "computed from {} character trees".format(
                  str(output_trees), str(len(character_dict))))
    else:
        print("Polymorphism removed. No polymorphism free "
              "informative subtree remaining.")
        sys.exit(1)

    return tree_dict


def rep_detector(character_dict):
    """
    Function to detect if repeated leaves are present in a dictionary of trees.

    """

    for t in character_dict.keys():
        cardl = list(Tree.get_leaf_names(t))
        cards = set(Tree.get_leaf_names(t))
        if len(cardl) != len(cards):
            return True

    return False


def triplet_extraction(infile, taxa_replacement_file=False):
    """
    Extract a dictionnary of triplets (keys) and weights (values) from a 
    prefix.triplet file generated by main_tripdec. Can use a taxa bloc 
    generated by main_tripdec for taxa name replacement.

    Parameters
    ----------
    infile : str
        Path of file containing triplets and weights.
    taxa_replacement_file : str, optional
        Path of a table file containing two columns. The first column
        corresponds to the names of the terminals of the newick stored in
        infile which must be integers, and the second column corresponds to
        their names the user wants to obtain at the end.
        All separators accepted. Example:
             1 Diceras
             2 Valletia
             3 Monopleura
        The default is False (no replacement).
    Returns
    -------
    triplet_dict : dict
        Dictionary of triplets (keys) and weights (values).

    """

    print("Loading triplet set")

    triplet_dict = {}

    infile = os.path.expanduser(infile)

    if not os.path.isfile(infile):
        print("ERROR: The file '" + infile + "' does not exist." +
                       "\nOperation aborted.")
        sys.exit(1)

    if taxa_replacement_file:
        taxa_dict = {}
        # detect separator in taxa_replacement_file
        with open(taxa_replacement_file, "r") as taxa_table1:
            try:
                dialect = csv.Sniffer().sniff(taxa_table1.read())
            except:
                print("ERROR: Could not determine separator in the "
                                  + "file '" + taxa_replacement_file
                                  + "'. The table is probably broken."
                                  + "\nOperation aborted.")
                sys.exit(1)

        # build taxanames to id converter
        with open(taxa_replacement_file, "r") as taxa_table1:
            data = csv.reader(taxa_table1, delimiter=dialect.delimiter)
            taxa_table2 = list(data)

            rowlist = [[i for i in row if i] for row in taxa_table2]

            for idtax, nametax in rowlist:
                taxa_dict[int(idtax)] = nametax

    # read triplet file and build triplet dictionary
    with open(infile, "r") as file_tree:
        for line in file_tree:
            if line.strip():
                tripletstr = line.split("    ")
                newstr = ''.join(
                    (ch if ch in '0123456789.-e' else ' ')
                    for ch in tripletstr[0])
                taxaint = [int(i) for i in newstr.split()]

                if taxa_replacement_file:
                    for i in [0,1,2]:
                        try:
                            taxa_dict[taxaint[i]]
                        except KeyError:
                            print("ERROR: The name '"
                                           + str(taxaint[i])
                                           + "' does not exists in the table "
                                           + "file '"
                                           + taxa_replacement_file
                                           + "'\nOperation aborted.")
                            sys.exit(1)

                    convert_trip = triplet({taxa_dict[taxaint[1]],
                                            taxa_dict[taxaint[2]]},
                                           {taxa_dict[taxaint[0]]})
                    triplet_dict[convert_trip] = Fraction(tripletstr[1])

                else:
                    trip = triplet({taxaint[1], taxaint[2]}, {taxaint[0]})
                    triplet_dict[trip] = Fraction(tripletstr[1])

    print("{} triplets loaded".format(str(len(triplet_dict))))

    return triplet_dict


def picklemerge(namelist):
    """
    Take as input a list of pickle file names, extract the triplet dictionaries
    and merge them.

    Parameters
    ----------
    namelist : list
        List of file names.

    Returns
    -------
    superdict : dict
        Dictionary of triplets (keys) and weights (values).

    """

    # load first pickle (first component triplet set)
    with open(namelist[0], 'rb') as pickle_file:
        superdict = defaultdict(int, pickle.load(pickle_file))

    # load each pickle file and add triplets to dict / cumulative weights
    for i in range(1, len(namelist)):
        with open(namelist[i], 'rb') as pickle_file:
            tripletdict_temp = pickle.load(pickle_file)
            for trip, FW in tripletdict_temp.items():
                superdict[trip] += FW

    return superdict


def tripdecFW(triplet_output=dict(), total_taxaset=False, character=Tree()):
    """
    Decompose a single tree into its triplets and compute their weights using
    the fractional weighting following Rineau et al. (2021).

    Rineau, V., Zaragüeta, R., & Bardin, J. (2021). Information content of
    trees: three-taxon statements, inference rules and dependency.
    Biological Journal of the Linnean Society, 133(4), 1152-1170.

    Parameters
    ----------
    triplet_output : dict, optional
        Used for recursion. The default is dict().
    total_taxaset : TYPE, optional
        Used for recursion. The default is False.
    character : ete3.Tree, optional
        A tree to decompose into triplets. The default is Tree().

    Returns
    -------
    dict
        Dictionary of triplets (keys) and weights (values).

    """

    if not total_taxaset:
        total_taxaset = set(character.get_leaf_names())

    # Compute triplet dictionary
    tree_tripdic = defaultdict(int)  # dict of triplets from tree
    nodaltripletdict = defaultdict(int)  # dict of triplets from component
    children_generator = (child_node for child_node in character.get_children(
        ) if not child_node.is_leaf())  # compute descendants info. nodes

    # Compute triplets from descendant nodes
    for child_node in children_generator:
        newtripletdict = tripdecFW(defaultdict(int),
                                   total_taxaset,
                                   character=child_node)

        for trip, FW in newtripletdict.items():
            tree_tripdic[trip] += FW

    # Compute triplets if not is not root
    if character.is_root():  # if node is root, end of analysis

        return tree_tripdic  # end

    else:
        taxa_in = set(character.get_leaf_names())  # taxa inside the node
        taxa_out = total_taxaset - taxa_in  # taxa outside the node
        totalnodeweight = len(taxa_out)*(len(taxa_in)-1)

        for taxa_in1, taxa_in2 in combinations(taxa_in, 2):  # all leaf couples
            for singleout in taxa_out:  # all sister taxa
                if not triplet({taxa_in1, taxa_in2},
                               {singleout}) in tree_tripdic:
                    nodaltripletdict[triplet({taxa_in1,
                                              taxa_in2},
                                             {singleout})] = 0

        for trip in nodaltripletdict.keys():
            tree_tripdic[trip] += Fraction(totalnodeweight,
                                           len(nodaltripletdict))

        return tree_tripdic


def tripdec(weighting, character):
    """
    Decompose a single tree into its triplets and compute their weights using
    a weighting scheme between:

        * Fractional weighting from Nelson and Ladiges (1992)
        * Uniform weighting from Nelson and Ladiges (1992)
        * Minimal weighting from Wilkinson et al. (2004)
        * Additive weighting : the weight of a triplet in additive weighting
        corresponds to the number of trees in which the triplet is present.
        * No weighting (all triplets have a weight of 1).

        Nelson, G., & Ladiges, P. Y. (1992). Information content and fractional
        weight of three-item statements. Systematic biology, 41(4), 490-494.

        Wilkinson, M., Cotton, J. A., & Thorley, J. L. (2004). The information
        content of trees and their matrix representations.
        Systematic Biology, 53(6), 989-1001.

    Parameters
    ----------
    weighting : str
        Weighting scheme to use between: * FWNL (Fractional weighting from
                                               Nelson and Ladiges),
                                         * UW (Uniform weighting from Nelson
                                               and Ladiges)
                                         * MW (Minimal Weighting),
                                         * AW (Additive Weighting),
                                         * NW (No Weighting).
    character : ete3.Tree
        A tree to decompose into triplets. The default is Tree().

    Returns
    -------
    dict
        Dictionary of triplets (keys) and weights (values).

    """

    triplet_output = dict()
    cardinal_character1 = set(Tree.get_leaf_names(character))  # leaf names
    internal_nodes1 = (node for node in character.traverse(
        strategy="preorder") if not node.is_leaf() and not node.is_root())
    tree_tripdic = defaultdict(int)
    card = len(cardinal_character1)
    MW_node = Fraction(6, card * (card - 1))

    for node in internal_nodes1:
        taxa_in = set(Tree.get_leaf_names(node))  # taxa inside the node
        taxa_out = cardinal_character1 - taxa_in  # taxa outside the node
        FW_node = Fraction(2, len(taxa_in))

        # build triplets from the node
        for taxa_in1, taxa_in2 in combinations(taxa_in, 2):
            for singleout in taxa_out:
                if weighting == "FWNL":
                    tree_tripdic[triplet({taxa_in1, taxa_in2},
                                         {singleout})] += FW_node

                elif weighting == "UW":
                    tree_tripdic[triplet({taxa_in1, taxa_in2},
                                         {singleout})] += 1

                if weighting == "MW":
                    tree_tripdic[triplet({taxa_in1, taxa_in2},
                                         {singleout})] = MW_node

                elif weighting == "AW" or weighting == "NW":
                    tree_tripdic[triplet({taxa_in1, taxa_in2},
                                         {singleout})] = 1

    if weighting == "NW":  # if no weighting

        triplet_output.update(tree_tripdic)

        return triplet_output

    else:
        for trip in tree_tripdic.keys():
            if trip in triplet_output:  # add new triplets
                triplet_output[trip] += tree_tripdic[trip]
            else:  # add weight if triplet already exists
                triplet_output[trip] = tree_tripdic[trip]

        return triplet_output


def tripdec_allweights(weighting, character):
    """
    Function to be used in parallel_tripdec only for multiprocessing
    computation.
    Decompose a single tree into its triplets and compute their weights
    using a weighting scheme between:

        * Fractional weighting from Rineau et al. (2021)
        * Fractional weighting from Nelson and Ladiges (1992)
        * Uniform weighting from Nelson and Ladiges (1992)
        * Minimal weighting from Wilkinson et al. (2004)
        * Additive weighting : the weight of a triplet in additive weighting
        corresponds to the number of trees in which the triplet is present.
        * No weighting (all triplets have a weight of 1).

        Nelson, G., & Ladiges, P. Y. (1992). Information content and fractional
        weight of three-item statements. Systematic biology, 41(4), 490-494.

        Rineau, V., Zaragüeta, R., & Bardin, J. (2021). Information content of
        trees: three-taxon statements, inference rules and dependency.
        Biological Journal of the Linnean Society, 133(4), 1152-1170.

        Wilkinson, M., Cotton, J. A., & Thorley, J. L. (2004). The information
        content of trees and their matrix representations.
        Systematic Biology, 53(6), 989-1001.

    Parameters
    ----------
    weighting : str
        Weighting scheme to use between: * FW (Fractional weighting from
                                               Rineau et al. 2021)
                                         * FWNL (Fractional weighting from
                                               Nelson and Ladiges),
                                         * UW (Uniform weighting from
                                               Nelson and Ladiges (1992)
                                         * MW (Minimal Weighting),
                                         * AW (Additive Weighting),
                                         * NW (No Weighting).
    character : ete3.Tree
        A tree to decompose into triplets. The default is Tree().

    Returns
    -------
    pickle_dict : dict
        Dictionary of triplets (keys) and weights (values) used for parallel
        computation in parallel_tripdec.
    """

    if not weighting in ("FW", "FWNL", "UW", "MW", "AW", "NW"):
        print("ERROR: " + weighting +
                       " is not a correct weighting method.\n" +
                       "Allowed weightings: FW, FWNL, MW, UW, AW, NW" +
                       "\nOperation aborted.")
        sys.exit(1)

    pickle_dict = defaultdict(int)
    tree_id = str(uuid.uuid4())

    cardinal_character1 = set(Tree.get_leaf_names(character))  # leaf names
    internal_nodes1 = (node for node in character.traverse(
        strategy="postorder") if not node.is_leaf() and not node.is_root())
    card = len(cardinal_character1)
    MW_node = Fraction(6, card * (card - 1))

    for node in internal_nodes1:
        pickle_name = str(uuid.uuid4()) + "_agatta.pickle"
        nodaltripletdict = dict()
        taxa_in = set(Tree.get_leaf_names(node))  # taxa inside the node
        taxa_out = cardinal_character1 - taxa_in  # taxa outside the node
        FWNL_node = Fraction(2, len(taxa_in))
        tempw = sum((len(t1)*len(t2) for t1, t2 in combinations((
            child_node.get_leaf_names() for child_node in node.get_children(
                )), 2)))
        FW_node = Fraction(len(taxa_out)*(len(taxa_in)-1), len(taxa_out)*tempw)

        for singleout in taxa_out:

            # generate all exclusive triplets from component
            if weighting in ("FW", "MW", "AW", "NW"):

                for taxalist1, taxalist2 in combinations((
                        child_node.get_leaf_names(
                        ) for child_node in node.get_children()), 2):
                    for taxa_in1, taxa_in2 in product(taxalist1, taxalist2):

                        if weighting == "FW":
                            nodaltripletdict[triplet(
                                in_taxa={taxa_in1, taxa_in2},
                                out_taxa={singleout})] = FW_node

                        elif weighting == "MW":
                            nodaltripletdict[triplet(
                                in_taxa={taxa_in1, taxa_in2},
                                out_taxa={singleout})] = MW_node

                        else:  # if additive weighting or no weighting
                            nodaltripletdict[triplet(
                                in_taxa={taxa_in1, taxa_in2},
                                out_taxa={singleout})] = 1

            # generate all triplets from component
            elif weighting in ("FWNL", "UW"):

                for taxa_in1, taxa_in2 in combinations(taxa_in, 2):

                    if weighting == "FWNL":
                        nodaltripletdict[triplet(
                            in_taxa={taxa_in1, taxa_in2},
                            out_taxa={singleout})] = FWNL_node

                    else:  # UW
                        nodaltripletdict[triplet(
                            in_taxa={taxa_in1, taxa_in2},
                            out_taxa={singleout})] = 1

        with open(pickle_name, 'wb') as pickle_file:
            pickle.dump(nodaltripletdict, pickle_file,
                        protocol=pickle.HIGHEST_PROTOCOL)

        pickle_dict[pickle_name] = tree_id

    return pickle_dict


def standard_tripdec(character_dict, weighting, dec_detail=False, 
                     prefix=False, verbose=True):
    """
    Compute triplets and weights from several rooted trees.
    Each tree is decomposed one after the other.

    Parameters
    ----------
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    weighting : str
        Weighting scheme to use between: * FW (Fractional weighting from
                                               Rineau et al. 2021)
                                         * FWNL (Fractional weighting from
                                               Nelson and Ladiges),
                                         * UW (Uniform weighting from
                                               Nelson and Ladiges 1992)
                                         * MW (Minimal Weighting),
                                         * AW (Additive Weighting),
                                         * NW (No Weighting).
    dec_detail : bool, optional
        If true, save a detailed table in csv format named 
        prefix.triplet_table.csv of triplet weights per character in addition 
        to the .triplet file. The default is False.
    prefix : str, optional
        Prefix of the saving file. The default is False.
    verbose : bool, optional
        Activate verbose mode if true. The default is True.

    Returns
    -------
    triplet_output : dict
        Dictionary of triplets (keys) and weights (values).

    """

    triplet_output = dict()
    triplet_output_per_char = dict()

    if weighting == "FW":

        for treedec, charnb in character_dict.items():
            triplet_output2 = tripdecFW(triplet_output,
                                        False, character=treedec)

            if dec_detail:
                triplet_output_per_char[charnb] = triplet_output2

            for trip, FW in triplet_output2.items():
                if trip in triplet_output:  # new triplets
                    triplet_output[trip] += FW
                else:  # sum triplet weights
                    triplet_output[trip] = FW

    elif weighting in ("FWNL", "MW", "AW", "UW", "NW"):
        for treedec, charnb in character_dict.items():
            triplet_output2 = tripdec(weighting, treedec)
            
            if dec_detail:
                triplet_output_per_char[charnb] = triplet_output2

            for trip, FW in triplet_output2.items():
                if trip in triplet_output and not weighting == "NW":
                    triplet_output[trip] += FW
                else:  # sum triplet weights
                    triplet_output[trip] = FW

    else:
        print("ERROR: " + weighting +
                       " is not a correct weighting method.\n" +
                       "Allowed weightings: FW, FWNL, MW, UW, AW, NW" +
                       "\nOperation aborted.")
        sys.exit(1)

    if dec_detail:
        
        # create detailed character table
        chartable = pd.DataFrame(0.000,
                         index=[str(trip) for trip in triplet_output.keys()], 
                         columns=triplet_output_per_char.keys())
        
        # fill table
        for charnb, triplet_output2 in triplet_output_per_char.items():
            for trip, FW in triplet_output2.items():
                
                chartable.loc[str(trip), charnb] += FW
                
        chartable.loc['Total'] = chartable.sum(axis=0)
        chartable.loc[:,'Total'] = chartable.sum(axis=1)
        chartable = chartable.round(decimals = 3)
        
        #save table
        chartable.to_csv(prefix + '.triplet_table.csv')
        print('Detailed triplet/character table computed.')
                    
    if prefix:
        with open(prefix+".triplet", "w") as tdfile:
            taxa_dict, code_dict, taxa_convers = taxa_to_numbers(
                                                                character_dict)
            if weighting in ("FW", "FWNL", "MW"):
                for trip, FW in triplet_output.items():

                    in_taxa = list(trip.in_taxa)
                    tdfile.write("(" + str(taxa_dict[list(trip.out_taxa)[0]])
                                 + ",("
                                 + str(taxa_dict[in_taxa[0]]) + ","
                                 + str(taxa_dict[in_taxa[1]]) + ")):    "
                                 + str(FW) + "    "
                                 + str(round(float(FW), 4))
                                 + "\n")

            else:
                 for trip, FW in triplet_output.items():

                    in_taxa = list(trip.in_taxa)
                    tdfile.write("(" + list(trip.out_taxa)[0] + ",("
                                 + in_taxa[0] + "," + in_taxa[1] + "));    "
                                 + str(FW) + "\n")

        with open(prefix+".taxabloc", "w") as taxa_bloc_file:
            for taxa, code in taxa_dict.items():
                taxa_bloc_file.write("{}    {}\n".format(str(code), taxa))

    return triplet_output


def parallel_tripdec(character_dict, weighting, prefix=False, ncpu="auto"):
    """
    Compute triplets and weights from several rooted trees in parallel.

    Parameters
    ----------
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    weighting : str
        Weighting scheme to use between: * FW (Fractional weighting from
                                               Rineau et al. 2021)
                                         * FWNL (Fractional weighting from
                                               Nelson and Ladiges),
                                         * UW (Uniform weighting from
                                               Nelson and Ladiges 1992),
                                         * MW (Minimal Weighting),
                                         * AW (Additive Weighting),
                                         * NW (No Weighting).
    prefix : str, optional
        Prefix of the saving file. The default is False.
    ncpu : str, optional
        Number of cpu to be used. The default is "auto".

    Returns
    -------
    triplet_output : dict
        Dictionary of triplets (keys) and weights (values).

    """

    def split(a, n):
        """
            Sort in n lists a single list a of pickle names
        """
        k, m = divmod(len(a), n)
        return list(a[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))

    # number of cpu
    if ncpu == "auto":
        ncpu = cpu_count()
    else:
        ncpu = int(ncpu)

    tmp = "tmp_Agatta_picKles_" + str(uuid.uuid4())  # folder for pickles

    try:
        os.mkdir(tmp)
    except FileExistsError:
        shutil.rmtree(os.getcwd() + '/' + tmp, ignore_errors=True)
        os.mkdir(tmp)
    finally:
        os.chdir(tmp)

    # computing triplets for each tree in parallel
    treetuple = tuple(character_dict)  # tuple dict for multiprocessing

    # creating dictionary
    with Pool(ncpu) as pool:  # multiprocessing activation
        func = partial(tripdec_allweights, weighting)
        pickle_double = pool.map(func, treetuple)  # list of triplet dict

    # pickle dictionaries extraction and merging
    picknum = len([item for sublist in [tuple(dicname.keys(
        )) for dicname in pickle_double] for item in sublist])

    if ncpu > picknum:  # if more cpu than pickles, use minimal number of cpu
        ncpu = picknum
        pickletuple = split([item for sublist in [tuple(dicname.keys(
            )) for dicname in pickle_double] for item in sublist], ncpu)
    else:  # else, use the maximum number of cpu allowed
        pickletuple = split([item for sublist in [tuple(dicname.keys(
            )) for dicname in pickle_double] for item in sublist], ncpu)

    with Pool(ncpu) as pool:  # multiprocessing
        dictuple = pool.map(picklemerge, pickletuple)  # list of triplet dicts

    if weighting == "NW":  # unique weight of 1
        shared_dict = {t: 1 for t in dictuple[0]}
        for d in dictuple[1:len(dictuple)]:
            for trip in d:
                shared_dict[trip] = 1

    else:  # sum weights
        shared_dict = dictuple[0]
        for d in dictuple[1:len(dictuple)]:
            for trip, FW in d.items():
                shared_dict[trip] += FW

    # pickles deletion
    os.chdir('..')
    shutil.rmtree(os.getcwd() + '/' + tmp, ignore_errors=True)

    if prefix:
        with open(prefix+".triplet", "w") as tdfile:
            taxa_dict, code_dict, taxa_convers = taxa_to_numbers(
                                                                character_dict)
            if weighting in ("FW", "FWNL", "MW"):
                for trip, FW in shared_dict.items():

                    in_taxa = list(trip.in_taxa)
                    tdfile.write("(" + str(taxa_dict[list(trip.out_taxa)[0]])
                                 + ",("
                                 + str(taxa_dict[in_taxa[0]]) + ","
                                 + str(taxa_dict[in_taxa[1]]) + ")):    "
                                 + str(FW) + "    "
                                 + str(round(float(FW), 4))
                                 + "\n")

            else:
                 for trip, FW in shared_dict.items():

                    in_taxa = list(trip.in_taxa)
                    tdfile.write("(" + list(trip.out_taxa)[0] + ",("
                                 + in_taxa[0] + "," + in_taxa[1] + "));    "
                                 + str(FW) + "\n")

        with open(prefix+".taxabloc", "w") as taxa_bloc_file:
            for taxa, code in taxa_dict.items():
                taxa_bloc_file.write("{}    {}\n".format(str(code), taxa))

    return shared_dict


def main_tripdec(input_item, prefix, taxa_replacement, weighting, parallel, 
                 dec_detail, method, verbose):
    """
    Main function of Agatta for decomposing trees into triplets and compute
    triplet weights using multiprocessing or not. The input trees can be a
    dictionary containing ete3 trees as keys or the path(s) to input file(s)
    (https://vrineau.github.io/AgattaDocs/Input%20files.html).
    Decompose trees into their triplets and compute weights using a weighting
    scheme between:

        * Fractional weighting from Rineau et al. (2021)
        * Fractional weighting from Nelson and Ladiges (1992)
        * Uniform weighting from Nelson and Ladiges (1992)
        * Minimal weighting from Wilkinson et al. (2004)
        * Additive weighting : the weight of a triplet in additive weighting
        corresponds to the number of trees in which the triplet is present.
        * No weighting (all triplets have a weight of 1).

        Nelson, G., & Ladiges, P. Y. (1992). Information content and fractional
        weight of three-item statements. Systematic biology, 41(4), 490-494.

        Rineau, V., Zaragüeta, R., & Bardin, J. (2021). Information content of
        trees: three-taxon statements, inference rules and dependency.
        Biological Journal of the Linnean Society, 133(4), 1152-1170.

        Wilkinson, M., Cotton, J. A., & Thorley, J. L. (2004). The information
        content of trees and their matrix representations.
        Systematic Biology, 53(6), 989-1001.

    Parameters
    ----------
    input_item : string or dictionary
        can be a list of file paths (newick, nexus, hmatrix, 
        or a dictionary of trees.
    prefix : str
        Prefix of the file to save. If prefix is set to false, no file is
        saved. The default is False.
    taxa_replacement : str, optional
        Path of a table file containing two columns. The first column
        corresponds to the names of the terminals of the newick stored in
        infile, and the second column corresponds to their names the user wants
        to obtain at the end. All separators accepted. Example:
             AA Diceras
             AB Valletia
             AC Monopleura
        The default is False (no replacement).
    weighting : str
        Weighting scheme to use between: * FW (Fractional weighting from
                                               Rineau et al. 2021)
                                         * FWNL (Fractional weighting from
                                               Nelson and Ladiges),
                                         * UW (Uniform weighting from
                                               Nelson and Ladiges 1992),
                                         * MW (Minimal Weighting),
                                         * AW (Additive Weighting),
                                         * NW (No Weighting).
    parallel : str or int
        Option for choosing if the analysis is made using multiprocessing or
        not. This argument can be:
            * "no" if the user does not wan to use multiprocessing.
            * "auto" for automatic detection of the number of cpu available.
            * any integer corresponding to the number of cpu the user wants to
              allow to the analysis.
    dec_detail : bool
        If true, save a detailed output table in csv format named 
        prefix.triplet_table.csv of triplet weights per character in addition 
        to the .triplet file. 
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS". The default is "TMS".
    verbose : bool
        Verbose mode if true.

    Returns
    -------
    triplet_dict : dict
        Dictionary of triplets (keys) and weights (values) without doublons.
    """

    if type(input_item) == list:
        character_dict = wrapper_character_extraction(input_item, 
                                                      taxa_replacement,
                                                      prefix,
                                                      verbose)
    
    else:
        character_dict = input_item

    # remove automatically repetitions if detected (user message printed)
    if rep_detector(character_dict):              
        character_dict = del_replications_forest(character_dict,
                                                     method=method,
                                                     prefix=prefix,
                                                     verbose=verbose)


    if not character_dict:
        print("ERROR: no informative character. Operation aborted.")
        sys.exit(1)

    # compute triplet dictionary for each tree in parallel
    print("Starting triplet decomposition and " + weighting + " computation")
    start = time.time()

    if parallel == "no":
        triplet_dict = standard_tripdec(character_dict, weighting, dec_detail, 
                                        prefix, verbose)
    else:
        triplet_dict = parallel_tripdec(character_dict, weighting,
                                        prefix, parallel)

    end = time.time()
    time_cptr = time.strftime('%H:%M:%S', time.gmtime(end - start))
    print("elapsed time (triplet decomposition and weighting): {}".format(
                                                                time_cptr))
    print("Generated " + str(len(triplet_dict))+" triplets")

    return triplet_dict



# -*- coding: utf-8 -*-
"""

    Agatta: Three-item analysis Python package
    Contact: Valentin Rineau - valentin.rineau@sorbonne-universite.fr

    Agatta is a set of tools in the cladistic framework to perform
    three-item analysis and associated operations in cladistics in python.

    https://github.com/vrineau/agatta

    This code is under license GNU GPLv3

"""

from .ini import hmatrix
from .ini import standardisation
from .ini import taxa_extraction
from .ini import taxa_to_numbers
from .ini import character_extraction
from .ini import wrapper_character_extraction
from .ini import taxa_triplet_extraction
from .analysis import main_tripdec
from .analysis import rep_detector
from .analysis import triplet_extraction
from .analysis import del_replications_forest
from .interpret import RI
from .interpret import rcc
from .interpret import constrict
from .interpret import character_states_test
from .interpret import brlen
from .search import search_pipeline
import os
import sys
import time
import datetime
import warnings
import treeswift
from ete3 import Tree


def triplet_nexus_file(triplet_dict, character_dict, weighting, analysis,
                       prefix, nrep=1000, logfile=False):
    """
    Outputs a three-item analysis file in nexus format readable by PAUP*4 from
    a dictionary of triplets and their corresponding weights.

    Swofford, D. L. 2003. PAUP*. Phylogenetic Analysis Using Parsimony
    (*and Other Methods). Version 4. Sinauer Associates, Sunderland,
    Massachusetts.

    Parameters
    ----------
    triplet_dict : dict
        Dictionary of triplets (keys) and weights (values).
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    weighting : str
        Weighting scheme to use between FW, FWNL, MW, AW, NW.
    analysis : str
        Add a line in the nexus file to state the type of analysis between
        branch and bound ("bandb") or heuristic ("heuristic").
    prefix : str
        Prefix of the saving nexus file prefix.nex. The complete path can be
        used.
    nrep : int, optional
        Add a line in the nexus file to state the number of replicates in case
        of heuristic search. The default is 1000.
    logfile : bool, optional
        IF true, Add a line in the nexus file to compute a log file during the
        PAUP4 analysis. The default is False.

    Returns
    -------
    None.

    """

    # weights
    count_zeroweights = 0  # zero-weight triplets (round)

    if weighting != "NW":
        i = 1  # triplet identifier
        triplet_list_FW = str()  # weight lines
        deltriplets = dict()

        for trip, FW in triplet_dict.items():

            # weights management using maximum weight in PAUP4
            if weighting == "FW" or weighting == "FWNL" or weighting == "MW":
                if character_dict:
                    FWdec = len(str(len(character_dict)))
                else:
                    FWdec = 2
                nexw = round((float(FW)), 8 - FWdec)
                if nexw == 0:
                    deltriplets[trip] = 0  # if zero-weight
                    count_zeroweights += 1
                elif nexw.is_integer():
                    triplet_list_FW += str(int(FW))
                else:
                    triplet_list_FW += str(nexw)
            else:
                nexw = 1
                triplet_list_FW += str(int(FW))  # NW, AW, UW (integers)

            if nexw != 0:
                triplet_list_FW += ":"
                triplet_list_FW += str(i)
                triplet_list_FW += ", "
                i += 1

        for trip in deltriplets.keys():
            triplet_dict.pop(trip)

    if character_dict:
        taxa_list = taxa_extraction(character_dict)
    else:
        taxa_list = taxa_triplet_extraction(triplet_dict)

    # if not prefix:
    #     root = Tk()
    #     prefix = filedialog.asksaveasfilename(
    #         title="Save three-taxon analysis PAUP nexus file", filetypes=((
    #             "nexus files", "*.nex"), ("all files", "*.*")))
    #     root.withdraw()

    # Nexus file computation
    with open(prefix+".nex", "w") as nexus_file:
        nexus_file.write("#NEXUS")
        nexus_file.write("\nbegin data;")
        nexus_file.write("\nDimensions ntax={} nchar={};".format(
            str(1 + len(taxa_list)), str(len(triplet_dict))))
        nexus_file.write("\nFormat symbols=\"0 1\" missing=?;")
        nexus_file.write("\n")
        nexus_file.write("\nMatrix")
        nexus_file.write("\n")

        # matrix computation
        for taxa_name in taxa_list:
            nexus_file.write("\n"+taxa_name+" ")
            newline = ""

            # for each triplet
            for triplet_column in triplet_dict.keys():
                if (taxa_name not in
                        triplet_column.out_taxa | triplet_column.in_taxa):
                    newline += "?"
                elif taxa_name in triplet_column.in_taxa:
                    newline += "1"
                else:
                    newline += "0"

            nexus_file.write(newline)

        nexus_file.write("\nroot "+"0"*len(triplet_dict))  # all 0 root

        nexus_file.write("\n;")
        nexus_file.write("\n")
        nexus_file.write("\nend;")
        nexus_file.write("\n")
        nexus_file.write("\nBegin Paup;")

        # weights lines
        if weighting != "NW":
            nexus_file.write("\nwts")
            nexus_file.write("\n"+triplet_list_FW[:-2])

        nexus_file.write("\n;")
        nexus_file.write("\noutgroup root /only;")

        if logfile:
            nexus_file.write("\nlog file={}_paup.log;".format(prefix))

        nexus_file.write("\nset maxtrees = 1000 increase=auto;")

        if analysis == "bandb":
            nexus_file.write("\nbandb;")
        elif analysis == "heuristic":
            nexus_file.write("\nhsearch addseq=random nreps={};".format(nrep))

        nexus_file.write("\nroottrees;")
        nexus_file.write("\nsavetrees /file={}.tre format=newick;".format(
            prefix))

        if logfile:
            nexus_file.write("\nlog stop;")  # end of log save

        nexus_file.write("\n;")
        nexus_file.write("\n")
        nexus_file.write("\nend;")

        if count_zeroweights > 0:
            print("WARNING, " + count_zeroweights +
                  " zero weight triplets deleted due to rounding")


def triplet_tnt_file(triplet_dict, character_dict, weighting, analysis,
                     prefix, nrep=1000, logfile=False):
    """
    Outputs a three-item analysis file in TNT format (Goloboff et al. 2008)
    from a dictionary of triplets and their corresponding weights.

    Goloboff, P. A., Farris, J. S., & Nixon, K. C. (2008). TNT, a free program
    for phylogenetic analysis. Cladistics, 24(5), 774-786.

    Parameters
    ----------
    triplet_dict : dict
        Dictionary of triplets (keys) and weights (values).
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    weighting : str
        Weighting scheme to use between FW, FWNL, MW, AW, NW.
    analysis : str
        Add a line in the TNT file to state the type of analysis between
        branch and bound ("bandb") or heuristic ("heuristic").
    prefix : str
        Prefix of the saving TNT file prefix.tnt. The complete path can be
        used.
    nrep : int, optional
        Add a line in the TNT file to state the number of replicates in case
        of heuristic search. The default is 1000.
    logfile : bool, optional
        IF true, Add a line in the TNT file to compute a log file during the
        TNT analysis. The default is False.

    Returns
    -------
    None.

    """

    count_zeroweights = 0  # zero-weight triplets (round)

    if weighting != "NW":
        triplet_list_FW = "ccode "
        wfactor = 1000 / float(max(triplet_dict.values()))
        itrip = 1
        deltriplets = dict()

        # matrix computation
        if (weighting == "FW" or weighting == "FWNL" or weighting == "MW" or
                max(triplet_dict.values()) > 1000):  # fractional weights
            for triplet_column, FW in triplet_dict.items():  # for each triplet

                nexw = int(float(FW)*wfactor)
                if nexw == 0:
                    deltriplets[triplet_column] = 0  # if zero-weight triplet
                    count_zeroweights += 1
                else:
                    triplet_list_FW += " /{} ".format(str(nexw))
                    triplet_list_FW += str(itrip)

                    itrip += 1

        # if rescale weight not needed
        else:
            for triplet_column, FW in triplet_dict.items():  # for each triplet

                nexw = 1
                triplet_list_FW += " /{} ".format(str(FW))
                triplet_list_FW += str(itrip)

                itrip += 1

        triplet_list_FW += "*;"  # end of the ccode line
        for trip in deltriplets.keys():
            triplet_dict.pop(trip)

    # computation of the TNT file
    # if prefix is None:
    #     root = Tk()
    #     prefix = filedialog.asksaveasfilename(
    #         title="Save three-taxon analysis PAUP nexus file",
    #         filetypes=(("nexus files", "*.nex"), ("all files", "*.*")))
    #     root.withdraw()

    if character_dict:
        taxa_list = taxa_extraction(character_dict)
    else:
        taxa_list = taxa_triplet_extraction(triplet_dict)

    tntstring = "xread\n{} {}".format(
        str(len(triplet_dict)), (str(1 + len(taxa_list))))
    tntstring += "\n"

    # matrix computation
    for taxa_name in taxa_list:
        tntstring += "\n"+taxa_name+" "
        newline = ""

        # for each triplet
        for triplet_column in triplet_dict.keys():

            if (taxa_name not in
                    triplet_column.out_taxa | triplet_column.in_taxa):
                newline += "?"
            elif taxa_name in triplet_column.in_taxa:
                newline += "1"
            else:
                newline += "0"

        tntstring += newline

    tntstring += "\nroot "+"0"*len(triplet_dict)  # all zeros root line

    tntstring += "\n;"
    tntstring += "\n"

    # weights line
    if weighting != "NW":
        tntstring += "\n"+triplet_list_FW

    tntstring += "\n;"
    tntstring += "\n"

    if logfile:
        tntstring += "\nlog {}.tnt.log;".format(prefix)  # if log

    tntstring += "\ntaxname =;"

    if analysis == "bandb":
        tntstring += "\nienum;"

    elif analysis == "heuristic":
        tntstring += "\nhold 10000;"  # max trees
        tntstring += "\nmult=replic {};".format(str(nrep))  # heuristic
        tntstring += "\nbbreak=tbr;"  # swapping method
        tntstring += "\ncollapse [; collapse 1;"  # collapse null branches

    tntstring += "\nexport - {}.tre;".format(prefix)  # export results to file
    tntstring += "\nquit"

    if count_zeroweights > 0:
        print("WARNING: " + str(count_zeroweights) +
              " zero weight triplets deleted due to rounding")

    with open(prefix+".tnt", "w") as tnt_file:
        tnt_file.write(tntstring)


def triplet_tmc_file(triplet_dict, character_dict, prefix,
                     weighting="FW"):
    """
    Outputs a three-item analysis file readable by Triplet MaxCut
    (Sevillya et al., 2016) from a dictionary of triplets and their
    corresponding weights.

    Sevillya, G., Frenkel, Z., & Snir, S. (2016). Triplet MaxCut: a new toolkit
    for rooted supertree. Methods in Ecology and Evolution, 7(11), 1359-1365.

    Parameters
    ----------
    triplet_dict : dict
        Dictionary of triplets (keys) and weights (values).
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    prefix : str
        Prefix of the saving TNT file prefix.tnt. The complete path can be
        used.
    weighting : str
        Weighting scheme to use between FW, FWNL, MW, AW, NW.

    Returns
    -------
    None.

    """

    # if prefix is None:
    #     root = Tk()
    #     prefix = filedialog.asksaveasfilename(
    #         title="Save three-taxon analysis TMC file", filetypes=((
    #             "text files", "*.txt"), ("all files", "*.*")))
    #     root.withdraw()

    # file for translating taxa names to numbers
    taxa_list = taxa_extraction(character_dict)
    taxa_dict, taxa_convers = taxa_to_numbers(taxa_list)

    with open(prefix+"_taxa.txt", "w") as TMC_file:
        for convers_line in taxa_convers:
            TMC_file.write(convers_line+"\n")
        TMC_file.write("\n")

    # string file
    tmcstring = ''
    for triplet1 in triplet_dict.keys():

        in_tax_list = list(triplet1.in_taxa)
        (out_taxa_str,) = triplet1.out_taxa

        # if weights are fractions
        if weighting == "FW" or weighting == "FWNL" or weighting == "MW":
            if character_dict:
                FWdec = len(str(len(character_dict)))
            else:
                FWdec = 2

            tmcstr = taxa_dict[in_tax_list[0]] + ","
            + taxa_dict[in_tax_list[1]] + "|" + taxa_dict[out_taxa_str]
            + "[:" + str(round((float(triplet1.FW)), 8 - FWdec)) + "] "
            tmcstring += tmcstr

        # if weights are integers
        else:
            tmcstring += taxa_dict[in_tax_list[0]] + ","
            + taxa_dict[in_tax_list[1]] + "|" + taxa_dict[out_taxa_str]
            + "[:" + str(int(triplet1.FW)) + "] "

    # write TMC file
    with open(prefix+".tmc", "w") as TMC_file:
        TMC_file.write(tmcstring)


def triplet_wqfm_wtqmc_file(triplet_dict, prefix, weighting="FW", precision=7):
    """
    Outputs a three-item analysis file readable by wqfm (Mahbub et al., 2021)
    and wTREE-QMC (Han & Molloy, 2024) from a dictionary of triplets and their 
    corresponding weights. To allow a three-item analysis with quartets, 
    each quartet possesses a leaf which is the root.

    Han and Molloy, 2024, Improved robustness to gene tree incompleteness, 
    estimation errors, and systematic homology errors with weighted TREE-QMC, 
    bioRxiv, https://doi.org/10.1101/2024.09.27.615467.

    Mahbub, M., Wahab, Z., Reaz, R., Rahman, M. S., & Bayzid, M. (2021).
    wQFM: Highly Accurate Genome-scale Species Tree Estimation from Weighted
    Quartets. Bioinformatics.
    
    Parameters
    ----------
    triplet_dict : dict
        Dictionary of triplets (keys) and weights (values).
    prefix : str, optional
        Prefix of the saving wqfm file.
    weighting : str, optional
        Weighting scheme to use between FW, FWNL, MW, AW, NW.
        The default is "FW".
    precision : int, optional
        Number of decimals to keep when rounding fractional weights.
        Increase this value for better precision.

    Returns
    -------
    None.

    """

    # core wqfm or wTREE-QMC file
    with open(prefix, "w") as wqfm_file:
        for triplet1, FW in triplet_dict.items():

            # if weights are fractions
            if weighting == "FW" or weighting == "FWNL" or weighting == "MW":
                wqfm_file.write("((" + list(triplet1.out_taxa)[0]
                                + ",root),(" + list(triplet1.in_taxa)[0] + ","
                                + list(triplet1.in_taxa)[1] + ")); "
                                + str(round(float(FW),precision))  # precision
                                + "\n")

            # if weights are integers
            else:
                wqfm_file.write("((" + list(triplet1.out_taxa)[0]
                                + ",root),(" + list(triplet1.in_taxa)[0]
                                + "," + list(triplet1.in_taxa)[1] + ")); "
                                + str(round(FW)) + "\n")


def triplet_to_file(triplet_dict, character_dict, prefix, analysis="heuristic",
                    nrep=1000, logfile=True, software="paup", weighting="FW"):
    """
    Outputs a three-item analysis file readable by TNT (Goloboff et al. 2008),
    by PAUP* (Swofford, 2003), by WQFM (Mahbub et al. 2021), or by wTREE-QMC 
    (Han & Molloy, 2024) from a dictionary of triplets and their corresponding 
    weights.

    Goloboff, P. A., Farris, J. S., & Nixon, K. C. (2008). TNT, a free program
    for phylogenetic analysis. Cladistics, 24(5), 774-786.

    Han and Molloy, 2024, Improved robustness to gene tree incompleteness, 
    estimation errors, and systematic homology errors with weighted TREE-QMC, 
    bioRxiv, https://doi.org/10.1101/2024.09.27.615467.
    
    Mahbub, M., Wahab, Z., Reaz, R., Rahman, M. S., & Bayzid, M. (2021).
    wQFM: Highly Accurate Genome-scale Species Tree Estimation from Weighted
    Quartets. Bioinformatics.

    Swofford, D. L. 2003. PAUP*. Phylogenetic Analysis Using Parsimony
    (*and Other Methods). Version 4. Sinauer Associates, Sunderland,
    Massachusetts.

    Parameters
    ----------
    triplet_dict : dict
        Dictionary of triplets (keys) and weights (values).
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    prefix : str, optional
        Prefix of the saving file. The complete path can be
        used. If set to none, a window appears to choose the location and name
        of the prefix for the nexus file. The default is None.
    analysis : str
        Add a line in the nexus file to state the type of analysis between
        branch and bound ("bandb") or heuristic ("heuristic").
    nrep : int, optional
        Add a line in the file to state the number of replicates in case
        of heuristic search. The default is 1000.
    logfile : bool, optional
        IF true, Add a line to outputs a log file during the  analysis.
        The default is False.
    software : str, optional
        Choose the software to use between 'paup' and 'tnt'.
        The default is "paup".
    weighting : str
        Weighting scheme to use between FW, FWNL, MW, AW, NW.
    Returns
    -------
    None.

    """

    print("Computing {} file".format(software))

    start = time.time()

    if software == "paup":
        triplet_nexus_file(triplet_dict, character_dict, weighting, analysis,
                           prefix, nrep, logfile)

    elif software == "tnt":
        triplet_tnt_file(triplet_dict, character_dict, weighting, analysis,
                         prefix, nrep, logfile)

    # elif software == "tmc":
    #     triplet_tmc_file(triplet_dict, character_dict, prefix, weighting)

    elif software == "wqfm":
        triplet_wqfm_wtqmc_file(triplet_dict, prefix+".wqfm", weighting)
        
    elif software == "wtree-qmc":
        triplet_wqfm_wtqmc_file(triplet_dict, prefix+".wtqmc", weighting)

    end = time.time()
    time_cptr = time.strftime('%H:%M:%S', time.gmtime(end - start))
    print("elapsed time (output file computation): {}".format(time_cptr))


def convert(infilelist, infiletype, prefix, parallel="auto", weighting="FW",
            analysis="heuristic", taxa_replacement=False, nreplicates=1000,
            logfile=True, software="paup", dec_detail=False, method="TMS", 
            verbose=True):
    """
    Outputs a three-item analysis file readable by TNT (Goloboff et al. 2008),
    by PAUP* (Swofford, 2003), by WQFM (Mahbub et al. 2021), or by
    wTREE-QMC (Han & Molloy, 2024) from :

        * a dictionary of triplets and their corresponding
          weights, e.g., computed from main_tripdec.
        * a dictionary of rooted trees.
        * a path to a text file containing newick trees.

    Goloboff, P. A., Farris, J. S., & Nixon, K. C. (2008). TNT, a free program
    for phylogenetic analysis. Cladistics, 24(5), 774-786.
    
    Han and Molloy, 2024, Improved robustness to gene tree incompleteness, 
    estimation errors, and systematic homology errors with weighted TREE-QMC, 
    bioRxiv, https://doi.org/10.1101/2024.09.27.615467.

    Mahbub, M., Wahab, Z., Reaz, R., Rahman, M. S., & Bayzid, M. (2021).
    wQFM: Highly Accurate Genome-scale Species Tree Estimation from Weighted
    Quartets. Bioinformatics.

    Swofford, D. L. 2003. PAUP*. Phylogenetic Analysis Using Parsimony
    (*and Other Methods). Version 4. Sinauer Associates, Sunderland,
    Massachusetts.

    Parameters
    ----------
    infilelist : dict or list of str
        Dictionary of ete3. Tree in keys or dictionary of triplets in keys and
        weights in values or list of paths of file(s) containing newick trees.
    infiletype : str
        Type of infile argument. Can be 'triplets' if infile is a dictionary
        of triplets or 'trees' if infile is a path to a tree file or a
        dictionary of trees.
    prefix : str, optional
        Prefix of the saving file. The complete path can be
        used. If set to none, a window appears to choose the location and name
        of the prefix for the nexus file. The default is None.
    parallel : str or int
        Option for choosing if the analysis is made using multiprocessing or
        not. This argument can be:

            * "not" if the user does not wan to use multiprocessing.
            * "auto" for automatic detection of the number of cpu available.
            * any integer corresponding to the number of cpu the user wants to
              allow to the analysis.

        This argument is only used if infiletype is 'trees'.
    weighting : str
        Weighting scheme to use between FW, FWNL, MW, AW, NW.
    analysis : str
        Add a line in the nexus file to state the type of analysis between
        branch and bound ("bandb") or heuristic ("heuristic").
    taxa_replacement : str, optional
        Path of a table file containing two columns. The first column
        corresponds to the names of the terminals of the newick stored in
        infile, and the second column corresponds to their names the user wants
        to obtain at the end. All separators accepted. Example:
             AA Diceras
             AB Valletia
             AC Monopleura
        The default is False (no replacement).
    nrep : int, optional
        Add a line in the file to state the number of replicates in case
        of heuristic search. The default is 1000.
    logfile : bool, optional
        IF true, Add a line to outputs a log file during the  analysis.
        The default is False.
    software : str, optional
        Choose the software to use between 'paup', 'tnt', 'wqfm', 'wtree-qmc'.
        The default is "paup".
    dec_detail : bool, optional
        If true, save a detailed table in csv format named 
        prefix.triplet_table.csv of triplet weights per character in addition 
        to the .triplet file. The default is False.
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS". The default is "TMS".
    verbose : bool, optional
        Verbose mode if True. The default is True.
    Returns
    -------
    None.

    """
    
    if dec_detail and not parallel == "no":
        print("WARNING: Detailed csv table (--detailed_tripdec) cannot be "
              +"constructed in parallel mode. To build the table, please set "
              +"parallel=no")

    if infiletype == "triplets":
        triplet_dict = triplet_extraction(infilelist, taxa_replacement)
        character_dict = False

    elif infiletype == "trees":
        if type(infilelist) == list:
            character_dict = wrapper_character_extraction(infilelist, 
                                                          taxa_replacement,
                                                          prefix,
                                                          verbose)

        else:
            character_dict = infilelist

        triplet_dict = main_tripdec(infilelist, prefix, taxa_replacement,
                                    weighting, parallel, dec_detail, 
                                    method, verbose)

    triplet_to_file(triplet_dict, character_dict, prefix, analysis,
                    nreplicates, logfile, software, weighting)


def agatta_analysis(infilelist, software_path, software="paup",
                    taxa_replacement=False, method="TMS", weighting="FW",
                    parallel="auto", prefix="agatta_out", analysis="heuristic",
                    nrep=1000, rosetta=False, chartest=False, ri=False,
                    consensus=False, pdf_file=False, dec_detail=False, 
                    nsupport=False, verbose=True):
    """
    Main function of the Agatta python package. Allow to perform a three-item
    analysis (Nelson & Platnick, 1991), e.g., in the context of systematics
    phylogenetics, using hierarchical characters (Cao et al. 2007), or in
    cladistic biogeography.

    The analysis can be performed using a text file containing a hierarchical
    matrix (see https://vrineau.github.io/AgattaDocs/Input%20files.html for 
    informations about the format) or newick rooted trees.

    Repetitions (polymorphism) are automatically removed according to:

        * The original algorithm of Nelson and Ladiges (1996) designed in the
          paradigm of cladistic biogeography.

        * The algorithm of Rineau et al. (in prep) designed to construct
          subtrees without repeated leaves while minimising the loss of
          information in terms of triplets. This is the default mode.

    The following options are available for analysing the results:

        * Computation of a consensus tree (strict consensus or reduced
          cladistic consensus; Wilkinson 1994)

        * Character state test procedure for hierarchical characters (Cao 2008
          modified by Rineau (2017)
          to analyse if character states are synapomorphies or homoplasies.

        * Retention index computation for all characters
          (Kitching et al., 1998).
          
        * Node support for each resulting cladogram based on triplets (amount 
          weighted triplets compatible with each node).

          Cao, N. (2008). Analyse à trois éléments et anatomie du bois des
          Fagales Engl (Doctoral dissertation, Paris, Muséum national
          d'histoire naturelle).

          Cao, N., Zaragüeta Bagils, R., & Vignes-Lebbe, R. (2007).
          Hierarchical representation of hypotheses of homology.
          Geodiversitas, 29(1), 5-15.

          Kitching, I. J., Forey, P., Humphries, C., & Williams, D. (1998).
          Cladistics: the theory and practice of parsimony analysis.
          Oxford University Press.

          Nelson, G. J., & Ladiges, P. Y. (1996). Paralogy in cladistic
          biogeography and analysis of paralogy-free subtrees.
          American Museum novitates 3167.

          Nelson, G., & Platnick, N. I. (1991). Three‐taxon statements:
          a more precise use of parsimony?. Cladistics, 7(4), 351-366.

          Rineau, V. (2017). Un nouveau regard cladistique sur l'anatomie
          comparée, la phylogénie, la systématique et la paléoécologie des
          rudistes (Bivalvia, Hippuritida) (Doctoral dissertation,
          Université Pierre et Marie Curie).

          Rineau, V., Moncel, M-H., & Zeitoun, V. (in prep). Revealing
          evolutionary patterns behind homogeneity: the case of the Paleolithic
          assemblages from Notarchirico (Southern Italy).

          Wilkinson, M. (1994). Common cladistic information and its consensus
          representation: reduced Adams and reduced cladistic consensus trees
          and profiles. Systematic Biology, 43(3), 343-368.

    Parameters
    ----------
    infilelist : list of str
        Paths of the file containing:

            * a hierarchical matrix (see 
              https://vrineau.github.io/AgattaDocs/Input%20files.html
              for informations about the format)
            * a single newick tree on each line. Example:

            (a,b,(c,d,(e,f)));
            ((c,e),(a,(b,(d,f))));

        The default is False (a selection window appears in this case).
    software_path : TYPE
        Path of the software designated with the "software" argument.
    software : str, optional
        Choose the software to use between 'paup', 'tnt', 'wqfm', 
        and 'wtree-qmc'.
        The default is 'paup'.
    taxa_replacement : str, optional
        Path of a table file containing two columns. The first column
        corresponds to the names of the terminals of the newick stored in
        infile, and the second column corresponds to their names the user wants
        to obtain at the end. All separators accepted. Example:
             AA Diceras
             AB Valletia
             AC Monopleura
        The default is False (no replacement).
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS". The default is "TMS".
    weighting : str
        Weighting scheme to use between:

            * FW: Fractional weighting from Rineau et al. (2021)
            * FWNL: Fractional weighting from Nelson and Ladiges (1992)
            * UW (Uniform weighting from Nelson and Ladiges 1992)
            * MW: Minimal weighting from Wilkinson et al. (2004)
            * AW: Additive weighting : the weight of a triplet in additive
              weighting corresponds to the number of trees in which the
              triplet is present.
            * NW: No weighting (all triplets have a weight of 1).

        Nelson, G., & Ladiges, P. Y. (1992). Information content and fractional
        weight of three-item statements. Systematic biology, 41(4), 490-494.
        Rineau, V., Zaragüeta, R., & Bardin, J. (2021). Information content of
        trees: three-taxon statements, inference rules and dependency.
        Biological Journal of the Linnean Society, 133(4), 1152-1170.
        Wilkinson, M., Cotton, J. A., & Thorley, J. L. (2004). The information
        content of trees and their matrix representations.
        Systematic Biology, 53(6), 989-1001.

        The default is 'FW'.
    parallel : str or int
        Option for choosing if the analysis is made using multiprocessing or
        not. This argument can be:

            * "not" if the user does not wan to use multiprocessing.
            * "auto" for automatic detection of the number of cpu available.
            * any integer corresponding to the number of cpu the user wants to
              allow to the analysis.

    prefix : str, optional
        Prefix of the saving file. The complete path can be
        used. If set to none, a window appears to choose the location and name
        of the prefix for the nexus file. The default is 'agatta_out'.
    analysis : str
        Add a line in the nexus file to state the type of analysis between
        branch and bound ("bandb") or heuristic ("heuristic").
        The default is 'heuristic'.
    nrep : int, optional
        Add a line in the file to state the number of replicates in case
        of heuristic search. The default is 1000.
    biogeo_tab : str
        path of table with semicolon separators and two columns, taxa in
        the left column and corresponding areas in the right column.
        The default is False (no standardisation is made).
    chartest : bool, optional
        If True, write two files with the results of the character state test
        procedure. The default is False.
    ri : bool, optional
        If True, write a file with a global retention index and one retention
        index for each character. The default is False.
    consensus : bool or str, optional
        Computation of a consensus tree between:

            * strict: strict consensus
            * rcc: reduced cladistic consensus (Wilkinson 1994).
            * False: no cosnensus computed.

        The default is False.
    pdf_file : bool or str, optional
        Can be used to compute pdf files if the character state procedure.
        pdf_file argument is a path to a floder where Agatta will write one pdf
        file for each character state. The default is False (no pdf generated).
    dec_detail : bool
        If true, save a detailed output table in csv format named 
        prefix.triplet_table.csv of triplet weights per character in addition 
        to the .triplet file. 
    verbose : bool, optional
        Verbose mode if True. The default is True.

    Returns
    -------
    None.

    """

    print("Starting analysis")

    with open(prefix+".log", "w") as log_file:
        log_file.write("Agatta parameters\n\n")
        log_file.write("Current date and time : ")
        now = datetime.datetime.now()
        log_file.write(now.strftime("%Y-%m-%d %H:%M:%S"))
        log_file.write("\n")
        log_file.write("Input file paths: "+ ", ".join(infilelist)+"\n")
        log_file.write("Prefix: "+prefix+"\n")
        log_file.write("Taxa replacement file:"+str(taxa_replacement)+"\n")
        log_file.write("Method: "+method+"\n")

        if rosetta:
            log_file.write("Standardisation: yes\n")
            log_file.write("Table path for standardisation: "+rosetta+"\n")
        else:
            log_file.write("Standardisation: no\n")

        log_file.write("Software used: "+software+"\n")

        if software_path:
            log_file.write("Software path: "+software_path+"\n")
        else:
            log_file.write("Software path: no file path\n")

        log_file.write("Weighting triplets: "+weighting+"\n")
        log_file.write("Parallelisation: "+parallel+"\n")

        if software == "wqfm" or software == "wtree-qmc": # always heuristic
            log_file.write("Analysis: heuristic")
        elif analysis == "heuristic":
            log_file.write("Analysis: heuristic with {} replicates\n".format(
            str(nrep)))
        else:
            log_file.write("Analysis: " + analysis + "\n")

        log_file.write("Character states test: "+str(chartest)+"\n")

        if pdf_file:
            log_file.write("Character states test pdf results location: " +
                           str(pdf_file)+"\n")
        else:
            log_file.write("Character states test: no pdf required by user\n")

        log_file.write("Retention index: "+str(ri)+"\n")
        log_file.write("Consensus: "+str(consensus)+"\n")
        log_file.write("Verbose: "+str(verbose)+"\n")

    # if input is a matrix, use hmatrix to convert in a tree list
    character_dict = wrapper_character_extraction(infilelist, taxa_replacement,
                                                  prefix, verbose)

    # standardisation option
    if rosetta:

        #save character_dict
        character_dict = standardisation(character_dict,
                                         rosetta,
                                         prefix,
                                         verbose=verbose)

    # remove automatically repetitions if detected (user message printed)
    if rep_detector(character_dict):
        character_dict = del_replications_forest(character_dict,
                                                 method=method,
                                                 prefix=prefix,
                                                 verbose=verbose)

    prefix_path = os.path.join(os.getcwd(), prefix)

    # compute triplets and weights from the tree list and save a nex/tnt file
    convert(character_dict,
            "trees",
            prefix,
            parallel,
            weighting,
            analysis,
            taxa_replacement,
            nrep,
            True,  # logfile
            software,
            dec_detail,
            method,
            verbose)

    # three-item analysis using PAUP*, TNT, WQFM, or wTREE-QMC
    if software == "paup":
        prefix_end = prefix_path+".nex"
        search_pipeline(prefix_end,
                        software_path,  # except if software_path == False
                        "paup")

    elif software == "tnt":
        search_pipeline(prefix+".tnt",
                        software_path,
                        "tnt")

    elif software == "wqfm":
        search_pipeline(prefix+".wqfm",
                        software_path,
                        "wqfm", prefix)

    elif software == "wtree-qmc":
        search_pipeline(prefix+".wtqmc",
                        software_path,
                        "wtree-qmc", prefix)
        
    # extract trees from output file and delete root
    cladogram_dict = dict()
    i = 1

    with warnings.catch_warnings():  # rerooting func poorly tested
        warnings.filterwarnings("ignore", category=UserWarning)
        if software == "tnt":

            with open(prefix + ".tre", "r") as tntfile:
                tntfile2 = []
                newickstring = False
                for line in tntfile:
                    if "[&U] \n" in line or "[&R] \n" in line:
                        tntfile2.append(line[:-2])  #remove tnt \n
                        newickstring = True
                    elif newickstring:
                        tntfile2.append(line.replace(" ",""))
                        newickstring = False
                    else:
                        tntfile2.append(line)
            with open(prefix + ".tre", "w") as tntfile:
                for line in tntfile2:
                    tntfile.write(line)

            tstreelist = treeswift.read_tree_nexus(prefix + ".tre")

        elif (software == "paup" or software == "wqfm" or 
              software == "wtree-qmc"):
            try:
                tstreelist = treeswift.read_tree_newick(prefix + ".tre")
            except RuntimeError:
                print("ERROR: " + software + " failed to find a tree.")
                if prefix[0] == "_" and software == "paup":
                    print("PAUP* doesn't like prefix names starting with _")
                sys.exit(1)

        if not isinstance(tstreelist, list):
            tstreelist = [tstreelist]

        for tstree in tstreelist:
            if isinstance(tstree, dict):
                for idt, tst in tstree.items():
                    nodedict = tst.label_to_node(selection='leaves')
                    tst.reroot(nodedict["root"])
                    tst.suppress_unifurcations()
                    newickstring = tst.newick().replace("root","")

                    if (newickstring.startswith("[&R] ") or
                        newickstring.startswith("[&U] ")):
                        cladogram_dict[Tree(
                            newickstring.split(" ")[1])] = idt
                    else:
                        cladogram_dict[Tree(newickstring)] = idt

            else:
                nodedict = tstree.label_to_node(selection='leaves')
                try:
                    tstree.reroot(nodedict["root"])
                except KeyError:
                    print("ERROR: A problem occurred when rooting. "
                          + "The problem may be due to old output files left " 
                          + "in the working directory. Please try to delete"
                          + " all files in the folder other than input files "
                          + "and restart the analysis. ")
                    
                    sys.exit(1)
                tstree.suppress_unifurcations()
                newickstring = tstree.newick().replace("root","")

                if (newickstring.startswith("[&R] ") or
                    newickstring.startswith("[&U] ")):
                    cladogram_dict[Tree(newickstring.split(" ")[1])] = i
                else:
                    cladogram_dict[Tree(newickstring)] = i
                i += 1

    # branch length computation
    if nsupport:
        brlen(prefix, cladogram_dict)
        
    # output file
    with open(prefix + ".tre", "w") as result_file:
        for tree in cladogram_dict.keys():
            result_file.write(tree.write(format=6) + "\n") #with support values

    if len(cladogram_dict.keys()) == 1:
        print("1 optimal tree found")
    else:
        print(str(len(cladogram_dict.keys())) + " optimal trees found")
        
    # consensus computation
    if consensus:

        # strict consensus
        if consensus == "strict":

            consensus = constrict(list(cladogram_dict.keys()), 
                                       prefix=False)
            
            # branch length computation
            if nsupport:
                brlen(prefix, consensus)
                
            # save consensus
            with open(prefix+".constrict", "w") as constrictfile:
                constrictfile.write(consensus.write(format=6)+"\n")

        # reduced cladistic consensus (Wilkinson, 1994)
        elif consensus == "rcc":

            consensus = rcc(list(cladogram_dict.keys()), prefix=False)
            
            # branch length computation
            if nsupport:
                brlen(prefix, consensus)
                      
            # save rcc
            with open(prefix+".rcc", "w") as rccfile:
                first_line = True
                i = 1
                for profiletree in consensus:
                    if first_line:
                        rccfile.write("Strict consensus:    "
                                      + profiletree.write(format=6)+"\n")
                        first_line = False
                    else:
                        rccfile.write("Profile tree " + str(i) + ":    "
                                      + profiletree.write(format=6) + "\n")
                        i += 1
            

    # character states test on the strict consensus
    if chartest:
        character_states_test({constrict(list(cladogram_dict.keys()),
                                         silent=True): 1},
                              character_dict,
                              prefix,
                              pdf_file)

    # retention index
    if ri:
        RI(cladogram_dict,
           character_dict,
           weighting=weighting,
           prefix=prefix)

    print('The analysis ended successfully')



# -*- coding: utf-8 -*-

"""
Agatta: Three-item analysis Python package
Contact: Valentin Rineau - valentin.rineau@sorbonne-universite.fr

Agatta is a set of tools in the cladistic framework to perform
three-item analysis and associated operations in cladistics in python.

https://github.com/vrineau/agatta

Usage:
    agatta analysis         <file>...        [options]
    agatta tripdec          <file>...        [options]
    agatta convert          <file>...        [options]
    agatta fp               <file>...        [options]
    agatta consensus        <file>           [options]
    agatta describetree     <file>           [options]
    agatta hmatrix          <file> ...       [options]
    agatta support          <file> <file>... [options]
    agatta chartest         <file> <file>... [options]
    agatta nri              <file> <file>... [options]
    agatta standardisation  <file> <file>... [options]
    agatta help             <command>
    agatta -h | --help
    agatta --version

Options:
    -h --help             Show general help information
    -s                    Silent mode
    -v                    Verbose mode
    --analysis=<type>     Type of searching method [default: heuristic]
    --nsupport            Compute triplet node support on reconstructed trees
    --chardec             Decompose trees into components
    --chartest            Compute hierarchical character states test
    --consensus=<type>    Specify type of consensus
    --directory=<dir>     Directory for output chartest files [default: ./]
    --detailed_tripdec    Compute a detailed table for triplet decomposition
    --filetype=<type>     Choose a tree/triplet/hmatrix file [default: trees]
    --index=<type>        Specify type of index to use [default: ri]
    --log                 Add a string to save an analysis log file
    --repetitions=<type>  Specify how to remove repetitions [default: TMS]
    --replicates=<int>    Number of replicates for an analysis [default: 1000]
    --parallel=<type>     Number of parallel cores to use [default: auto]
    --pdf                 Save a pdf file for chartest or nri results
    --prefix=<name>       Prefix of the output file [default: agatta_out]
    --ri                  Calculate Retention Index (Kitching et al. 1998)
    --rosetta=<path>      Path of the file for taxa conversion
    --showtaxanames       Print the list of terminals
    --softpath=<path>     Path of the software used
    --software=<type>     Software used for the pipeline [default: paup]
    --taxarep1=<path>     Table for leaves replacement for file 1
    --taxarep2=<path>     Table for leaves replacement for file 2
    --weighting=<type>    Specify the type of triplet weighting [default: FW]
    --rnri_codes=<path>   Compute NRI with a split of characters given in csv       
    --rnri_totaltree      Compute a pdf with absolute triplet support by node
    --rnri_rescaling      Rescale values in rnri_totaltree

"""

from .ini import helper
from .ini import hmatrix_several
from .ini import checkargs
from .ini import standardisation
from .ini import character_extraction
from .ini import wrapper_character_extraction
from .analysis import main_tripdec
from .analysis import del_replications_forest
from .interpret import RI_path
from .interpret import triplet_distance
from .interpret import rcc
from .interpret import constrict
from .interpret import describe_forest
from .interpret import chartest
from .interpret import NRI
from .out import agatta_analysis
from .out import convert
from .__version__ import __version__
import os
import sys
import time
import docopt


def main():
    # docopt flags parsing
    arguments = docopt.docopt(__doc__, version=__version__)

    def coremain():

        start_time = time.time()

        print("Agatta {}".center(80).format(__version__))
        print("Three-item analysis Python package".center(80))
        print()

        # check parsing
        checkargs(arguments)

        # Complete analysis
        if arguments["analysis"]:
            agatta_analysis(arguments["<file>"],  # one or several input files
                            arguments["--softpath"],
                            arguments["--software"],
                            arguments.get("--taxarep1", False),
                            arguments["--repetitions"],
                            arguments["--weighting"],
                            arguments["--parallel"],
                            arguments["--prefix"],
                            arguments["--analysis"],
                            arguments["--replicates"],
                            arguments.get("--rosetta", False),
                            arguments["--chartest"],
                            arguments["--ri"],
                            arguments.get("--consensus", False),
                            arguments["--pdf"],
                            arguments.get("--detailed_tripdec", False),
                            arguments.get("--nsupport", False),
                            arguments.get("-v", False))

        # triplet decomposition
        elif arguments["tripdec"]:
            main_tripdec(arguments["<file>"],  # one or several input files
                         arguments["--prefix"],
                         arguments.get("--taxarep1", False),
                         arguments["--weighting"],
                         arguments["--parallel"],
                         arguments.get("--detailed_tripdec", False),
                         arguments["--repetitions"],
                         arguments.get("-v", False))

        # retention index calculation
        elif arguments["support"]:
            # retention index
            if arguments["--index"] == "ri":

                RI_path(arguments["<file>"][0],
                   arguments["<file>"][1:],  # one or several input files
                   arguments.get("--taxarep1", False),
                   arguments.get("--taxarep2", False),
                   arguments["--repetitions"],
                   arguments["--weighting"],
                   arguments["--prefix"]+".txt")

            # inter-tree retention index
            elif arguments["--index"] == "tripdistance":  # only two trees

                triplet_distance(list(character_extraction(
                    arguments["<file>"][0],
                    arguments.get("--taxarep1", False),
                    verbose=False).keys())[0],
                    list(character_extraction(
                        arguments["<file>"][1],
                        arguments.get("--taxarep2", False),
                        verbose=False).keys())[0],
                    arguments["--prefix"],
                    arguments["--weighting"])

        # character states testing procedure
        elif arguments["chartest"]:
            chartest(arguments["<file>"][0],
                     arguments["<file>"][1:],  # one or several input files
                     arguments.get("--taxarep1", False),
                     arguments.get("--taxarep2", False),
                     arguments["--repetitions"],
                     arguments["--prefix"],
                     arguments["--pdf"],
                     arguments.get("-v", False))

        # nodal retention index
        elif arguments["nri"]:
            NRI(arguments["<file>"][0],
                     arguments["<file>"][1:],  # one or several input files
                     arguments.get("--taxarep1", False),
                     arguments.get("--taxarep2", False),
                     arguments["--prefix"],
                     arguments.get("--rnri_codes", False),
                     arguments["--weighting"],
                     arguments["--repetitions"],
                     arguments.get("--rnri_totaltree", True),
                     arguments.get("--rescaling", False),
                     arguments["--pdf"])

        # convert
        elif arguments["convert"]:
            convert(arguments["<file>"],   # one or several input files
                    arguments["--filetype"],
                    arguments["--prefix"],
                    arguments["--parallel"],
                    arguments["--weighting"],
                    arguments["--analysis"],
                    arguments.get("--taxarep1", False),
                    arguments["--replicates"],
                    arguments.get("--log", False),
                    arguments["--software"],
                    False,
                    arguments["--repetitions"],
                    arguments.get("-v", False))

        # free-subtree paralogy analysis
        elif arguments["fp"]:
            del_replications_forest(wrapper_character_extraction(
                             arguments["<file>"],  # one or several input files
                             arguments.get("--taxarep1", False),
                             prefix=arguments["--prefix"],
                             verbose=arguments.get("-v", False)),
                             method=arguments["--repetitions"],
                             prefix=arguments["--prefix"],
                             verbose=arguments.get("-v", False))

        # consensus
        elif arguments["consensus"]:
            # reduced cladistic consensus
            if arguments["--consensus"] == "rcc":

                rcc(list(character_extraction(
                    arguments["<file>"][0],  # only one file
                    arguments.get("--taxarep1", False)).keys()),
                    arguments["--prefix"])

            # strict consensus
            else:
                constrict(list(character_extraction(
                          arguments["<file>"][0],   # only one file
                          arguments.get("--taxarep1", False)).keys()),
                          arguments["--prefix"])

        # describe trees
        elif arguments["describetree"]:   # only one file
            describe_forest(character_extraction(arguments["<file>"][0], 
                                                 info_tree=False),
                            arguments["--prefix"],
                            arguments.get("--showtaxanames", False))

        # standardisation
        elif arguments["standardisation"]:
            standardisation(arguments["<file>"][1:],  # one or several input
                              arguments["<file>"][0],
                              arguments["--prefix"],
                              verbose=arguments.get("-v", False))

        # transform hierarchical matrix into a tree list
        elif arguments["hmatrix"]:
            hmatrix_several(arguments["<file>"],  # one or several input
                    arguments["--prefix"],
                    arguments["--chardec"],
                    arguments["-v"])

        elif arguments["help"]:
            helper(arguments["<command>"])

        # display elapsed time
        elapsed_time = time.time() - start_time
        time_cptr = time.strftime("%H:%M:%S", time.gmtime(elapsed_time))
        print("elapsed time (total): {}".format(time_cptr))

    # class silent mode
    class HiddenPrints:
        def __enter__(self):
            self._original_stdout = sys.stdout
            sys.stdout = open(os.devnull, 'w')

        def __exit__(self, exc_type, exc_val, exc_tb):
            sys.stdout.close()
            sys.stdout = self._original_stdout

    # choice silent mode or not
    if arguments["-s"]:
        with HiddenPrints():
            coremain()
    else:
        coremain()
